import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.beans.*;
import javax.swing.plaf.basic.BasicInternalFrameUI;
import java.util.*;

/**
 * This is the ClientFrame that controls all aspects of the application
 * @author Jung Son
 * @version 1.0 final release
 */ 
public class ClientFrame extends JFrame implements Observer {
  private static final int WIDTH = 800;
  private static final int HEIGHT = 700;
  
  private LoginScreen login;
  private WaitScreen wait;
  private PlayScreen play;
  private ClientNetwork cn;
  private int game_mode = ClientModel.IN_WAITROOM;
  private String user_name;
  private JDesktopPane desktop= new JDesktopPane();
  
  /**
   * Construct and initialize ClientFrame
   */ 
  public ClientFrame() {
    setTitle("OpenSpace Interactive v1.0");
    setSize(WIDTH, HEIGHT);
    setResizable(false);
    
    // create new desktop pane object where internal jframes are going to be stored
    setContentPane(desktop);
    displayLogin(desktop);
  }
  
  /**
   * Creates and displays the log-in screen
   * @param desktop The component to add the log-in screen
   */ 
  public void displayLogin(JDesktopPane desktop) {
    // create login screen
    login = new LoginScreen(this);
    login.addJDesktop(desktop);
  }
  
  /**
   * Creates and displays the wait screen
   * @param cf The ClientFrame class
   * @param cn The ClientNetwork class
   * @param user_name The user name of the application user
   */ 
  public void displayWait(ClientFrame cf, ClientNetwork cn, String user_name) {
    game_mode = ClientModel.IN_WAITROOM;
    wait = new WaitScreen(cf, cn);
    wait.addJDesktop(desktop);
    this.user_name = user_name;
    setTitle("OpenSpace Interactive v1.0 - [" + user_name + "] Connected");
    cn.sendMsg(ClientModel.GAME_INFO + "_\r");
  }
  
  /**
   * Creates and displays the play screen
   * @param cf The ClientFrame class
   * @param cn The ClientNetwork class
   * @param game_mode The current game mode
   */ 
  public void displayPlay(ClientFrame cf, ClientNetwork cn, int game_mode) {
    this.game_mode = game_mode;
    play = new PlayScreen(cf, cn,game_mode);
    play.addJDesktop(desktop);
  }
  
  /**
   * Establish observer connection with the client network class
   * @param cn The ClientNetwork class
   */ 
  public void setConnection(ClientNetwork cn) {
    this.cn = cn; 
    this.cn.addObserver(this);
  }
  
  /**
   * Get the user name
   * @return The user name
   */ 
  public String getUserName() {
    return user_name;
  }
  
  /**
   * Change the game mode
   * @param gm The new game mode to change to
   */ 
  public void setGameMode(int gm) {
    game_mode = gm;
  }

  /**
   * Handle messages sent by the server
   * @param o The observable object
   * @param arg An argument passed to the notifyObservers method
   */ 
  public void update(Observable o, Object arg) {
    String msg = cn.getMsg();
    String[] loc = msg.split("_");
    int key = Integer.parseInt(loc[0]);
    
    //System.out.println(msg);
    
    if (game_mode == ClientModel.ILLUSTRATOR || game_mode == ClientModel.ASSISTANT) {
      switch(key) {
        case ClientModel.ERROR: {
          switch(Integer.parseInt(loc[1])) {
            default: {
              JOptionPane.showMessageDialog(null, "Error", "Error.ID "+ loc[1], JOptionPane.WARNING_MESSAGE );
              break;
            }
          }
          break;
        }
        case ClientModel.CHAT: {
          play.userTalked(loc[1], loc[2]);
          break;
        }
        case ClientModel.JOIN: {
          if (game_mode == ClientModel.ILLUSTRATOR) {
            play.updatePlayerList(loc[1], play.findBlankSpot());
            String playersInfo = ClientModel.ADD_USER + "_";
              JTextField[] playerList = play.getPlayerList();
            
            for (int i=0; i<6;  i++) {
              if (playerList[i].getText().length() != 0) {
                playersInfo = playersInfo + playerList[i].getText() + "_" + i + "_";
              }
            }
            cn.sendMsg(playersInfo + user_name + "_\r");   
            play.indicateIllustrator(user_name);
            
            String welcome = loc[1] + " has joined the room";
            play.userTalked("mod", welcome);
            cn.sendMsg(ClientModel.CHAT + "_mod_" + welcome + "_\r"); 
          }
          break;
        }
        case ClientModel.ADD_USER: {
          for (int i=1; i<loc.length-1; i+=2) {
            play.updatePlayerList(loc[i], Integer.parseInt(loc[i+1]));
          }
          play.indicateIllustrator(loc[loc.length-1]);
          break; 
        }
        case ClientModel.LEAVE_GAME: {
            JTextField[] playerList = play.getPlayerList();
            for (int i=0; i<6; i++) {
              if ((playerList[i].getText()).equals(loc[1])) {
                play.updatePlayerList("", i);
                play.userTalked("mod", loc[1]+" has left the room");
              }
            }
          break;
        }
        case ClientModel.START: {
          play.startTimer();
          break;
        }       
        case ClientModel.STOP: {
          play.stopTimer();
          break;
        }
        case ClientModel.NEW_ILLUSTRATOR: {
          if (loc[1].compareTo(user_name) == 0) {
            play.setGameMode(ClientModel.ILLUSTRATOR);
            game_mode = ClientModel.ILLUSTRATOR;
            play.updateDrawingPanel(this, play, cn);
          } else {
            play.updateDrawingPanel(this, play, cn);
          }
          play.indicateIllustrator(loc[1]);
          break;
        }
        case ClientModel.REQUEST_TOPIC: {
          if (game_mode == ClientModel.ILLUSTRATOR) {
            play.displayTopic(loc[1]);
            System.out.println(loc[1]);
          } else {
            String hint = "";
            for (int i=0; i<loc[1].length();i++) {
              hint = hint+"*";
            }
            play.displayTopic(hint);
          }
          break;
        }
      }
    } else if (game_mode == ClientModel.IN_WAITROOM) {
      switch (key) {
        case ClientModel.GAME_INFO: {
          if (loc.length == 1) {
            wait.updatePlayRoomList("none");
          } else {
            wait.updatePlayRoomList(msg.substring(loc[0].length()+1, msg.length()));
          }
          break;
        }
        // you are authorized to join.
        case ClientModel.LOGIN: {
          displayWait(this, cn, loc[1]);
          login.dispose();
          break;
        }
        case ClientModel.JOIN: {
          game_mode = ClientModel.ASSISTANT;
          displayPlay(this, cn, ClientModel.ASSISTANT);
          wait.dispose();
          break;
        }
        case ClientModel.CREATE: {
          game_mode = ClientModel.ILLUSTRATOR;
          displayPlay(this, cn, ClientModel.ILLUSTRATOR);
          wait.dispose();
          break;
        }
        case ClientModel.ERROR: {
          switch(Integer.parseInt(loc[1])) {
            case ClientModel.LOGIN: {
              JOptionPane.showMessageDialog(null, "User name is already taken. Please select a new user name and try again.", "Error.ID "+ loc[1], JOptionPane.WARNING_MESSAGE );
              break;
            }
            case ClientModel.JOIN: {
              JOptionPane.showMessageDialog(null, "You cannot join this room. Please select another.", "Error.ID "+ loc[1], JOptionPane.WARNING_MESSAGE );
              break;
            }
          }
          break;
        }
      }
    }
  }
}