import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import javax.swing.*;
import java.io.*;
import java.awt.image.BufferedImage;

/**
 * This is the paint board of the application. All the drawing takes place here
 * @author Jung Son
 * @version 1.0 final release
 */ 
public class DrawingCanvas extends JPanel implements Observer {
  
  private Color penColor = Color.BLACK;
  private Color prevPenColor = Color.BLACK;
  private int penSize;
  private ClientNetwork cn;
  private String user_name;
  
  private int prevX = 0;
  private int prevY = 0;
  private int currX;
  private int currY;

  private BufferedImage _bufImage = null;
  private int game_mode; 
  private Boolean enableDraw = false;
  private Boolean eraseSelected = false;
  
  /**
   * Construct and initialize the paint board
   * @param cn The ClientNetwork class
   * @param user_name The user name of the application user
   * @param game_mode The game mode of the game
   */ 
  public DrawingCanvas(final ClientNetwork cn, String user_name, int game_mode) {
    setBackground(Color.WHITE);
    
    this.cn = cn;
    this.user_name = user_name;
    this.game_mode = game_mode;

    setBackground(Color.WHITE);
    this.repaint();
  }
  
  /**
   * Allow the user to start drawing
   */ 
  public void startDrawing() {
    addMouseListener(new MouseAdapter() {
      public void mouseReleased(MouseEvent event) {
        myMouseReleasedHandler();
      }
    });
    
    addMouseMotionListener(new MouseMotionAdapter() {
      public void mouseDragged(MouseEvent event) {
        myMouseDragPaint(event);
      }
    });
  }
  
  /**
   * Stop any further drawing by the user
   */ 
  public void disableDrawing() {
    for (int i=0; i<getMouseListeners().length; i++) {
      removeMouseListener(getMouseListeners()[i]);
    }
    
    for (int i=0; i<getMouseMotionListeners().length; i++) {
      removeMouseMotionListener(getMouseMotionListeners()[i]);
    }
  }
  
  /**
   * User is draggin the mouse and drawing on the canvas. Show what the user is drawing
   * @param event
   */ 
  private void myMouseDragPaint(java.awt.event.MouseEvent event) {
    currX = event.getX();
    currY = event.getY();
    
    //Previous coordinates will only be 0,0 if this is a new line.
    //If this is the case, set previous coords equal to new coords
    //which will begin a new line.
    if (prevX == 0) { 
      prevX = currX;
    }
    if (prevY == 0) {
      prevY = currY; 
    }

    Graphics2D grafarea = _bufImage.createGraphics();
    drawCurrentShape(grafarea);

    this.repaint();
    cn.sendMsg(ClientModel.DRAW + "_" + prevX + "_" + prevY + "_" + currX + "_" + currY + "_" + penSize + "_" + penColor.getRGB() + "_\r");
    
    prevX = currX;
    prevY = currY;
  }
  
  /**
   * User has stopped drawing for now. Reset conditions
   */ 
  private void myMouseReleasedHandler() {
    //Reset coordinates to 0,0 indicating that the line has been broken
    prevX = 0;
    prevY = 0;
  }
  
  /**
   * Display the current shape drawn by the user on the screen
   * @param g2 The Graphics g2 object
   */ 
  private void drawCurrentShape(Graphics2D g2) {
    BasicStroke stroke = new BasicStroke((float)penSize, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
    g2.setStroke(stroke);
    g2.setColor(penColor);
    g2.drawLine(prevX, prevY, currX, currY);
  }
  
  /**
   * Calls the UI delegate's paint method, if the UI delegate is non-null. We pass the delegate a copy of the Graphics 
   * object to protect the rest of the paint code from irrevocable changes (for example, Graphics.translate). 
   * @param g The Graphics object to protect
   */ 
  public void paintComponent(Graphics g) {
    super.paintComponent(g);
    Graphics2D g2 = (Graphics2D)g;  // downcast to Graphics2D

    if (_bufImage == null) {
      //this is the first time, initialize _bufImage
      int w = this.getWidth();
      int h = this.getHeight();
      _bufImage = (BufferedImage)this.createImage(w, h);
      Graphics2D gc = _bufImage.createGraphics();
      gc.setColor(Color.white);
      gc.fillRect(0, 0, w, h); // fill in background
    }
    g2.drawImage(_bufImage, null, 0, 0);  // draw previous shapes
    
    if (prevX != 0 && prevY != 0 ) {
      drawCurrentShape(g2);
    }
  }

  /**
   * If you are an illustrator, draw by listening to the server for paint coordinates
   * @param o The observable object
   * @param arg An argument passed to the notifyObservers method
   */ 
  public void update(Observable o, Object arg) {
      String msg = cn.getMsg();
      String[] loc = msg.split("_");
      int key = Integer.parseInt(loc[0]);
      
      if (key == ClientModel.START) {
        enableDraw = true;
      } else if (key == ClientModel.STOP) {
        enableDraw = false;
      }
      
      // if you are an assistant, draw by receiving coordinates from the server
      if (game_mode == ClientModel.ASSISTANT && enableDraw) {
        switch(key) {
          case ClientModel.DRAW: {
            if (loc[1] != null && loc[2] != null && loc[3] != null && loc[4] != null && loc[5] != null && loc[6] != null){
              prevX = Integer.parseInt(loc[1]);
              prevY = Integer.parseInt(loc[2]);
              currX = Integer.parseInt(loc[3]);
              currY = Integer.parseInt(loc[4]);
              penSize = Integer.parseInt(loc[5]);
              penColor = new Color(Integer.parseInt(loc[6]));
              Graphics2D grafarea = _bufImage.createGraphics();
              drawCurrentShape(grafarea);
              this.repaint();
            }
            break;
          }
          case ClientModel.ERASE: {
            clearScreen();
            break;
          }
        }
      } 
      // if you are an illustrator, listen to server for change in color
      else if (game_mode == ClientModel.ILLUSTRATOR && enableDraw) {
        if (key == ClientModel.COLOR_CHANGE) {
          setColor(new Color(Integer.parseInt(loc[1])));
        }
      }
  }
               
  
  /**
   * Change the color of the pen
   * @param c Color to change the pen
   */ 
  public void setColor(Color c) {
    if (!eraseSelected) {
      penColor = c;
    }
  }
  
  /**
   * Get current color used for the pen
   * @return current color
   */ 
  public Color getColor() {
    return penColor;
  }

  
  /**
   * Clear the paint board of any drawing
   */ 
  public void clearScreen() {
    myMouseReleasedHandler();
    _bufImage = null;
    this.repaint();
    
    if (game_mode == ClientModel.ILLUSTRATOR ) {
      cn.sendMsg(ClientModel.ERASE + "_\r");
    }
  }
  
  /**
   * Set the size of the pen
   * @param s The size of the pen
   */ 
  public void setSize(int s) {
    penSize = s;
  }
  
  /**
   * Set whether the eraser is selected or not
   * @param b Is the eraser selected
   */  
  public void setEraser(boolean b) {
    if (b) {
      eraseSelected = true;
      if (penColor != Color.WHITE) {
	prevPenColor = penColor;
      }
      penColor = Color.WHITE;
    } else {
      eraseSelected = false;
      penColor = prevPenColor;
    }
  }
  
  
}
