import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.beans.*;
import javax.swing.plaf.basic.BasicInternalFrameUI;
import javax.swing.BorderFactory; 
import javax.swing.border.EtchedBorder;

/**
 * Displays the play screen of the application 
 * @author Jung Son
 * @version 1.0 final release
 */ 
public class PlayScreen extends JInternalFrame {
  
  private ClientNetwork cn;
  private ClientFrame cf;
  private int game_mode;
  private DrawingPanel dp;
  private JButton send;
  
  private JTextArea chatDisplay = new JTextArea();
  private JTextField[] playerList = new JTextField[6];
  
  /**
   * Construct and initialize the play screen
   * @param cf The ClientFrame class
   * @param cn The ClientNetwork class
   * @param game_mode The game mode of the game
   */ 
  public PlayScreen(ClientFrame cf, final ClientNetwork cn, int game_mode) {
    super("", false, false, false, false);
    
    this.cn = cn;
    this.cf = cf;
    this.game_mode = game_mode;
    
    JPanel topPanel = new JPanel();
    getContentPane().add(topPanel, BorderLayout.NORTH);
    
    dp = new DrawingPanel(cf, this, cn);
    getContentPane().add(dp, BorderLayout.CENTER);
    displayGamePlayers();
    
    JPanel bottomPanel = new JPanel(new FlowLayout());
    // add chatbox
    createChatBox(bottomPanel);
    
    getContentPane().add(bottomPanel, BorderLayout.SOUTH);
    
    if (game_mode == ClientModel.ILLUSTRATOR) {
      updatePlayerList(cf.getUserName(), 0);
      indicateIllustrator(cf.getUserName());
    }
    
    show();
  }
  
  /**
   * Create locations where game players will be listed
   */ 
  private void displayGamePlayers() {
    GridBagLayout gridbag = new GridBagLayout();
    GridBagConstraints c = new GridBagConstraints();
    JPanel rightPanel = new JPanel(gridbag);
    JPanel leftPanel = new JPanel(gridbag);

    c.fill = GridBagConstraints.BOTH;
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.insets = new Insets(30,10,30,10);
    c.anchor = GridBagConstraints.NORTH;
    
    // display first half on the left panel and the other half on the right panel
    for (int i=0; i<playerList.length; i++) {
      if (i < 3) {
        addPlayers(leftPanel, i, gridbag, c);
      } else {
        addPlayers(rightPanel, i, gridbag, c);
      }
    }
    
    rightPanel.add(Box.createRigidArea(new Dimension(0,10)));
    leftPanel.add(Box.createRigidArea(new Dimension(0,10)));
    getContentPane().add(rightPanel, BorderLayout.EAST);
    getContentPane().add(leftPanel, BorderLayout.WEST);
  }
  
  /**
   * Set conditions for spot that a user is going to be shown in
   * @param panel The panel where ther user is going to be added
   * @param i Player's location number
   * @param gridbag Gridbag layout manager
   * @param c The contraints to be applied
   */
  private void addPlayers(JPanel panel, int i, GridBagLayout gridbag, GridBagConstraints c) {
    JTextField player = new JTextField(7);
    // set size of the player field
    Dimension size = new Dimension(80, 80);
    player.setMaximumSize(size);
    player.setMinimumSize(size);
    player.setPreferredSize(size);
    
    player.setEditable(false);
    player.setText("");
    player.setHorizontalAlignment(JTextField.CENTER ); 
    player.setBorder( BorderFactory.createEtchedBorder(EtchedBorder.RAISED));
    playerList[i] = player;
    gridbag.setConstraints(player, c);
    panel.add(player);
  }
  
  /**
   * Creates chat area where the users can talk
   * @param bottomPanel The panel to attach the chat area
   */ 
  private void createChatBox(JPanel bottomPanel) {
    JPanel chatBox = new JPanel();
    chatBox.setLayout(new BoxLayout( chatBox, BoxLayout.Y_AXIS));
    
    final JTextField chat_field = new JTextField(10);
    send = new JButton("Send Message");
    
    // add keyboard input to allow the user to simply press enter to submit
    chat_field.addKeyListener(new KeyListener() {
      public void keyPressed(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_ENTER) {
          send.doClick();
        }
      }
      public void keyTyped(KeyEvent e) {}
      public void keyReleased(KeyEvent e){}
    });
    chatBox.add(chat_field);
    chatBox.add(Box.createRigidArea(new Dimension(0,13)));
    
    // button to submit the chat
    send.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        String msg = chat_field.getText();
        if (msg != "" && msg.length() > 0) {
          userTalked(cf.getUserName(), msg);
          cn.sendMsg(ClientModel.CHAT + "_" + cf.getUserName() + "_" + msg + "_\r");
          chat_field.setText("");
        }
      }
    }); 
    
    chatBox.add(send);
    chatBox.setBorder( BorderFactory.createCompoundBorder(
                                                             BorderFactory.createCompoundBorder(
                                                                                                BorderFactory.createTitledBorder("Chat Box"),
                                                                                                BorderFactory.createEmptyBorder(5,5,5,5)),
                                                             chatBox.getBorder()));
    
    // set conditions for display area
    chatDisplay.setLineWrap(true);
    chatDisplay.setWrapStyleWord(true);
    chatDisplay.setEditable(false);
    
    // set size of the chat dialogue display area
    JScrollPane scrollPane = new JScrollPane(chatDisplay);
    scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
    scrollPane.setPreferredSize(new Dimension(550, 100));
    scrollPane.setMaximumSize(new Dimension(550, 100));
    scrollPane.setMinimumSize(new Dimension(550, 100));
    
    // create border for dialogue display
    scrollPane.setBorder( BorderFactory.createCompoundBorder(
                                                             BorderFactory.createCompoundBorder(
                                                                                                BorderFactory.createTitledBorder("Client Chat Dialogue"),
                                                                                                BorderFactory.createEmptyBorder(5,5,5,5)),
                                                             scrollPane.getBorder()));
    
    bottomPanel.add(scrollPane);
    bottomPanel.add(chatBox);
  }
  
  /**
   * Update the drawing panel of the play screen. Drawing Panel is where the paint board and art-related buttons are 
   * located
   * @param cf The ClientFrame class
   * @param ps The PlayScreen class
   * @param cn The ClientNetwork class
   */ 
  public void updateDrawingPanel(ClientFrame cf, PlayScreen ps, ClientNetwork cn) {
    remove(dp);
    dp = new DrawingPanel(cf, ps, cn);
    getContentPane().add(dp, BorderLayout.CENTER);
    revalidate();
    show();
  }
  
  /**
   * Display user chat messages
   * @param user The name of the user that sent the chat message
   * @param msg The content of the message
   */ 
  public void userTalked(String user, String msg) {
    chatDisplay.append(user + ": " + msg +"\n");
    
    if (game_mode == ClientModel.ILLUSTRATOR) {
      if (msg.equalsIgnoreCase(dp.getTopic())) {
        String temp = ": WINNER [" + user + "]    ANSWER [" + msg + "]    TIME [" + dp.getTimeElapsed() + " Sec]";
        chatDisplay.append("mod" + temp + "\n");
        cn.sendMsg(ClientModel.CHAT + "_mod_" + temp + "_\r");
        cn.sendMsg(ClientModel.STOP + "_\r");
      }
    }
    
    // automatically scroll to the end of the scrollpane
    chatDisplay.setCaretPosition(chatDisplay.getDocument().getLength());
  }
  
  /**
   * Indicate the illustrator by highlighting the illustrator box
   * @param name The name of the illustrator to highlight
   */ 
  public void indicateIllustrator(String name) {
      for (int i=0; i<6; i++) {
        if (playerList[i].getText().equals(name)) {
          playerList[i].setBackground(Color.LIGHT_GRAY  );
        } else {
          playerList[i].setBackground(null);
        }
      }
  }
  
  /**
   * Update the display of the players in the game room
   * @param str The name of the player to update
   * @param i The player's location number
   */ 
  public void updatePlayerList(String str, int i) {
    playerList[i].setText(str);
  }
  
  /**
   * Find where in the player display area is not occupied
   * @return The spot number of unoccupied area
   */ 
  public int findBlankSpot() {
    for (int i=0; i<6; i++) {
      if (playerList[i].getText().length() == 0) {
        return i;
      }
    }
    return ClientModel.ROOM_FULL;
  }
  
  /**
   * Counts the total number of players currently in a game room
   * @return The number of players in the room
   */ 
  public int getNumPlayers() {
    int counter = 0;
    for (int i=0; i<6; i++) {
      if (playerList[i].getText().length() > 0) {
        counter++;
      }
    }
    return counter;
  }
  
  /**
   * Change the game mode
   * @param game_mode The new game mode to change to
   */ 
  public void setGameMode(int game_mode) {
    this.game_mode = game_mode;
  }
  
  /**
   * Get the current game mode
   * @return The current game mode
   */ 
  public int getGameMode() {
    return game_mode;
  }
  
  /**
   * Get a list of all the players current in the game room
   * @return List of all players
   */ 
  public JTextField[] getPlayerList() {
    return playerList;
  }
  
  /**
   * Starts the timer - how long the game will last
   */ 
  public void startTimer() {
    dp.startTimer();
  }
  
  /**
   * Stops the timer - how long the game will last
   */ 
  public void stopTimer() {
    dp.stopTimer();
  }
  
  /**
   * Enable or disable chatting option
   * @param b True to enable chat, false to disable chat
   */ 
  public void enableChat(boolean b) {
    if (b) {
      send.setEnabled(true);
    } else {
      send.setEnabled(false);
    }
  }
  
  /**
   * Display the game topic to draw on the canvas
   * @param tpc The topic for the illustrator to draw
   */ 
  public void displayTopic(String tpc) {
    dp.displayTopic(tpc);
  }
  
  /**
   * Add the log-in screen to the specified container
   * @param desktop The component to add the log-in screen
   */ 
  public void addJDesktop(JDesktopPane desktop) {
    // disable border and frame top bar
    ((BasicInternalFrameUI)this.getUI()).setNorthPane(null);
    this.setBorder(null);
    
    desktop.add(this);
    
    // set internal screen to maximum size
    try {
      this.setMaximum(true);
    } catch(PropertyVetoException e){}; 
  }
}