import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;
import java.beans.*;
import javax.swing.plaf.basic.BasicInternalFrameUI;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import java.util.EventObject;

/**
 * Displays the wait screen of the application
 * @author Jung Son
 * @version 1.0 final release
 */
public class WaitScreen extends JInternalFrame{

  private ClientNetwork cn;
  private ClientFrame cf;
  private String create_title;
  private String join_title;

  private JTable table;
  private Object[][] cells;
  private JScrollPane scrollPane;

  /**
   * Construct and initialize the wait screen
   * @param cf The ClientFrame class
   * @param cn The ClientNetwork class
   */
  public WaitScreen(ClientFrame cf, ClientNetwork cn) {
    super("", false, false, false, false);

    this.cn = cn;
    this.cf = cf;

    createWaitroomButtons();

    String[] columnNames = {"Room Title", "Number of Users", "Status"};
    cells = new Object[0][3];
    table = new JTable(cells, columnNames);
    scrollPane = new JScrollPane(table);
    getContentPane().add(scrollPane, BorderLayout.CENTER);

    JPanel topPanel = new JPanel();
    ImageIcon logo = new ImageIcon("openspace.jpg");
    JLabel logoLabel = new JLabel(logo);
    topPanel.add(logoLabel);

    JPanel leftPanel = new JPanel();
    JPanel bottomPanel = new JPanel();

    // panels not defined yet
    getContentPane().add(topPanel, BorderLayout.NORTH);
    getContentPane().add(bottomPanel, BorderLayout.SOUTH);
    getContentPane().add(leftPanel, BorderLayout.WEST);

    show();
  }

  /**
   * Displays and updates the list of rooms available to join
   * @param gameList The list of rooms to shown on the display
   */
  public void updatePlayRoomList(String gameList) {
    getContentPane().remove(scrollPane);

    // titles of the table
    String[] columnNames = {"Room Title", "Number of Users", "Status"};

    if (!gameList.equals("none")) {
      String[] info = gameList.split("_");

      cells = new Object[info.length][3];
      for (int i=0; i<info.length; i++) {
        String[] detail = info[i].split("-");

        cells[i][0] = detail[0];
        cells[i][1] = detail[1] + "/6";

        if (detail[2].equals("y")) {
          cells[i][2] = "available";
        }else if (detail[2].equals("n")) {
          cells[i][2] = "in session";
        }
      }
    } else {
      cells = new Object[0][3];
    }

    table = new JTable(cells, columnNames) {
      public boolean isCellEditable(int rowIndex, int vColIndex) {
        return false;
      }
    };

    // listen for which rooom the user selects
    table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    ListSelectionModel rowSM = table.getSelectionModel();
    final javax.swing.table.TableModel model = table.getModel();
    rowSM.addListSelectionListener(new ListSelectionListener() {
      public void valueChanged(ListSelectionEvent e) {
        //Ignore extra messages.
        if (e.getValueIsAdjusting()) return;

        ListSelectionModel lsm = (ListSelectionModel)e.getSource();
        if (lsm.isSelectionEmpty()) {
          System.out.println("No rows are selected.");
        } else {
          int selectedRow = lsm.getMinSelectionIndex();
          join_title = (String)model.getValueAt(selectedRow, 0);
          System.out.println("Row " + selectedRow
                               + ", Room Title: " + join_title);
        }
      }
    });

    // add scroll bar
    scrollPane = new JScrollPane(table);
    getContentPane().add(scrollPane, BorderLayout.CENTER);

    // create cell renderer
    TableCellRenderer centerRenderer = new CenterRenderer();

    //  Use renderer on a specific column
    for (int i=0; i<3; i++) {
      TableColumn column = table.getColumnModel().getColumn(i);
      column.setCellRenderer( centerRenderer );
    }

    //  Use renderer on a specific Class
    table.setDefaultRenderer(String.class, centerRenderer);

    // update the view
    revalidate();
    show();
  }


  /**
   * Create buttons for the wait room
   */
  private void createWaitroomButtons() {
    JPanel rightPanel = new JPanel();
    GridBagLayout gridbag = new GridBagLayout();
    GridBagConstraints c = new GridBagConstraints();
    rightPanel.setLayout(gridbag);

    JButton create = new JButton("Create");
    JButton join = new JButton("Join");
    JButton refresh = new JButton("Refresh");
    JButton quit = new JButton("Quit");

    // create button - create new game
    create.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        create_title = JOptionPane.showInputDialog("Please Enter Room Title");
        if (create_title != null && create_title.length()  != 0) {
          cn.sendMsg(ClientModel.CREATE + "_" + create_title + "_\r");
        }
      }
    });

    // join button - join game
    join.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        cn.sendMsg(ClientModel.JOIN + "_" + join_title + "_\r");
      }
    });

    // refresh button - refresh the content of game room list
    refresh.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        cn.sendMsg(ClientModel.GAME_INFO + "_\r");
      }
    });

    // quit button - quit application
    quit.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent event) {
        cn.sendMsg(ClientModel.LOGOUT + "_\r");
        System.exit(0);
      }
    });

    // add all the displays to the wait room screen
    c.fill = GridBagConstraints.BOTH;
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.insets = new Insets(2,5,2,5);
    c.anchor = GridBagConstraints.NORTH;

    gridbag.setConstraints(create, c);
    rightPanel.add(create);
    gridbag.setConstraints(join, c);
    rightPanel.add(join);
    gridbag.setConstraints(refresh, c);
    rightPanel.add(refresh);
    gridbag.setConstraints(quit, c);
    rightPanel.add(quit);

    JPanel outerPanel = new JPanel(new BorderLayout());
    outerPanel.add(rightPanel, BorderLayout.NORTH);
    getContentPane().add(outerPanel, BorderLayout.EAST);
  }

  /**
   * Add the log-in screen to the specified container
   * @param desktop The component to add the log-in screen
   */
  public void addJDesktop(JDesktopPane desktop) {
    // disable border and frame top bar
    ((BasicInternalFrameUI)this.getUI()).setNorthPane(null);
    this.setBorder(null);

    desktop.add(this);

    // set internal screen to maximum size
    try {
      this.setMaximum(true);
    } catch(PropertyVetoException e){};
  }

  /*
   **  Center and disable the text editing in play room list
   */
  class CenterRenderer extends DefaultTableCellRenderer {
    public CenterRenderer() {
      setHorizontalAlignment( CENTER );
    }

    public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
      super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
      return this;
    }
  }
}