import java.util.*;

/**
 * This is the GameRoom state that keeps track of a game's information
 * and any messages that need to be relayed between clients in a game.
 * @author Calvin Chin and David Couvrette
 * @version 1.0 final
 */
public class GameRoom implements Room {

  private static int ROOM_SIZE = 6;

  // Array of all the game clients in this room.
  // NOTE: Currently need to resolve issues with handling the deletion of a user from the list.
  // Possible solution: Go back to hashmaps?
  private List inGameClients;
  private int illustrator;
  private int gameMode;
  private GameState state;
  private MainRoom main;

  /**
   * Construct the GameRoom
   * @param main The main room that created this game
   */
  public GameRoom(MainRoom main) {
    this.main = main;
    inGameClients = new ArrayList();
    illustrator = 0;
    state = new GameState();
  }

  /**
   * Construct the GameRoom and specify a Client as illustrator.
   * @param firstPlayer The client creating the room (illustrator)
   * @param main The main room that created this game
   */
  public GameRoom(Client firstPlayer, MainRoom main) {
    this(main);
    firstPlayer.setRoom(this);
    inGameClients.add(firstPlayer);
  }

  /**
   * Gets the number of players in the game room
   * @return The number of players in the game room
   */
  public int getNumPlayers() {
    return inGameClients.size();
  }

  /**
   * Adds a new client to the set of current clients
   * If the client is the first to be added to the room
   * that client is set as the first illustrator
   * Sends a message to each client in the game room
   * giving the name of the new client
   * format of message: 101_userName_
   * @param newClient The new client that is joining the server
   */
  public synchronized void addNewClient(Client newClient) {
    // if adding first player, set that player to illustrator
    if (inGameClients.size() == 0) {
      inGameClients.add(newClient);
      illustrator=0;
    }
    else {
      // Non hackish code begins
      if (inGameClients.size() < ROOM_SIZE) {
        inGameClients.add(newClient);
      } else {
        throw new IllegalArgumentException("This room is full.");
      }
      Iterator it = inGameClients.iterator();
      String msg = "101_" + newClient.getUserName() + "_";
      while (it.hasNext()) {
        Client user = (Client) it.next();
        if (!user.getUserName().equals(newClient.getUserName()))
          sendTo(user.getUserName(), msg);
      }
    }
    newClient.setRoom(this);
  }

  /**
   * Determines whether a client can join the game room.
   * checks that the room is not full and game is not started
   * @return true if the user can join, otherwise false
   */
  public boolean canJoin() {
    return inGameClients.size() < ROOM_SIZE &&
      state.getGameMode() != GameState.IN_GAME;
  }

  /**
   * Sends a given message to a given user
   * @param userName The user name
   * @param msg The message being sent
   * @return true if the message succeeded, otherwise false
   */
  public boolean sendTo(String userName, String msg) {
    Client c = (Client)inGameClients.get(findUser(userName));
    return c.sendMsg(msg);
  }

  /**
   * Sends a given message to all clients in the room
   * except for the sender
   * @param sender The client who sent the message
   * @param msg The message to be sent
   */
  public synchronized void sendToAllOthers(Client sender, String msg) {
    Iterator it = inGameClients.iterator();
    while (it.hasNext()) {
      Client user = (Client) it.next();
      if (!(user == sender))
        sendTo(user.getUserName(), msg);
    }
  }

  /**
   * Sends a message to all clients.
   * @param message The message that is sent to every client
   */
  public synchronized void sendToAllClients(String message) {
    Iterator it = inGameClients.iterator();
    while (it.hasNext())
      ((Client) it.next()).sendMsg(message);
  }

  /**
   * Remove a client from the room.
   * Note: Currently very hackish. Need to resolve the issue of empty slots
   * by going to hashmaps again, or making a repack() method.
   * Will be running tests to see which is faster - Calvin.
   * @param userName The name of the client that is being removed
   * @return The client that has just been removed
   */
  public synchronized Client removeClient(String userName) {
    return (Client) inGameClients.remove(findUser(userName));
  }

  /**
   * Handles Messages from the client
   * @param msg the message or command that is being sent
   * @param sender the client who sent the message
   */
  public void handleMsg(String msg, Client sender) {
    int tag = MessageParser.getTag(msg);
    String message = MessageParser.getMsg(msg);
    //System.out.println(msg);
    switch (tag) {
      case Room.START:
        startGame(sender);
        break;
      case Room.STOP:
        state.endGame();
        sendToAllClients(msg);
        break;
      case Room.LEAVE_GAME:
        leaveGame(sender);
        break;
      case Room.DRAW:
        sendToAllOthers(sender, msg);
        break;
      case Room.ERASE:
        sendToAllOthers(sender, msg);
        break;
      case Room.COLOR_CHANGE:
        sendToAllOthers(sender, msg);
        break;
      case Room.SIZE_CHANGE:
        sendToAllOthers(sender, msg);
        break;
      case Room.TOPIC_REQUEST:
        getTopic();
        break;
      case Room.CHAT:
        sendToAllOthers(sender, msg);
        break;
      case Room.ADD_USER:
        sendToAllOthers(sender, msg);
        break;
      /*case Room.REMOVE_USER:
        sendToAllOthers(sender, msg);
        break;*/
      case Room.NEW_ILLUSTRATOR:
        sendToAllOthers(sender, msg);
        break;
      default:
        System.out.println("GameRoom: Unknown message type. Message follows: " + msg);
        break;
    }
  }

  /**
   * Starts the game and sends the client
   * a confirmation that the game is starting
   * @param sender The client that is starting the game
   */
  private void startGame(Client sender) {
    // send topic to illustrator
    // tell everyone else the game is starting
    if (!(GameState.IN_GAME == state.getGameMode())) {
      sendToAllClients("5_");
      state.startGame();
    }
    else {
      sender.sendMsg("0_5_");
    }
  }

  private void leaveGame(Client sender) {
    // is this the last player leaving?
    if (inGameClients.size() == 1) {
      main.removeGame(this);
    }
    else {
      // is the illustrator leaving?
      // or only 2 players left?
      // if so, end the game, and set illustrator to 0
      if (findUser(sender.getUserName()) == illustrator ||
          inGameClients.size() < 3) {
        illustrator = 0;
        state.endGame();
      }
    }
    sendToAllOthers(sender, "12_" + sender.getUserName());
    main.addNewClient((Client)inGameClients.remove(findUser(sender.getUserName())));
  }

  private void getTopic() {
    sendToAllClients(Room.TOPIC_REQUEST + "_" + state.getTopic() + "");
  }

  /**
   * Sends the update to all other clients
   */
  private void update(String msg, Client sender) {
    sendToAllOthers(sender, msg);
  }

  /**
   * Sends the update to just the sender
   */
  private void updateSender(String msg, Client sender) {
    sendTo(sender.getUserName(),msg);
  }

  /**
   * Sends the update to just the illustrator
   */
  private void updateIllustrator(String msg, Client sender) {
    sendTo(getIllustrator().getUserName(),msg);
  }

  /**
   * Sends the update to all the assistants
   */
  private void updateAssistants(String msg, Client sender) {
    String illustrator = getIllustrator().getUserName();
    Iterator it = inGameClients.iterator();
    while (it.hasNext()) {
      String user = (String) it.next();
      if (!user.equals(illustrator))
        sendTo(user, msg);
    }
  }

  /**
   * Checks to see if a given user is in the room.
   * @param userName the user name to search for
   * @return true if the room contains the given user name, otherwise false
   */
  public boolean containsUser(String userName) {
    return findUser(userName) >= 0;
  }

  /**
   * Looks through the array for a user. Returns -1 if no user found.
   * @param userName User name to search for.
   */
  public int findUser(String userName) {
    Iterator it = inGameClients.iterator();
    int i = 0;
    while (it.hasNext()) {
      Client t = (Client)it.next();
      if (t.getUserName().equals(userName))
        return i;
      i++;
    }

    // TODO: Switching this to throw an exception instead.
    return -1;
  }

  /**
   * Returns the current game mode.
   */
  public int getGameMode() {
    return gameMode;
  }

  /**
   * Returns the currrent illustrator.
   */
  public Client getIllustrator() {
    return (Client)inGameClients.get(illustrator);
  }
}