import java.util.*;

/**
 * Main Room for Users
 * This is the main control for the server.
 * It holds a collection of clients in the main room and
 * available games to connect to.
 * @author David Couvrette
 * @version 1.0 final release
 */
public class MainRoom implements Room {

  // hashmap to store all the clients that are in the main server room
  private HashMap clients;
  // hashmap to store all the current games on this server
  private HashMap games;


  /**
   * Construct and initialize the server
   */
  public MainRoom() {
    clients = new HashMap();
    games = new HashMap();
    ScanIncomingConnections scan = new ScanIncomingConnections(this);
    Thread t = new Thread(scan);
    t.start();
  }

  /**
   * Adds a new client to the set of current clients
   * Starts a new thread that waits for another client
   * @param newClient The new client that is joining the server
   */
  public synchronized void addNewClient(Client newClient) {
    if (clients.containsKey(newClient.getUserName()))
      throw new IllegalArgumentException("That name is already taken");
    clients.put(newClient.getUserName(), newClient);
    newClient.setRoom(this);
  }

  /**
   * Sends a given message to a given client
   * @param userName The name of the user to send the message to
   * @param msg The message that is being sent
   */
  public synchronized boolean sendTo(String userName, String msg) {
    Client c = (Client)clients.get(userName);
    return c.sendMsg(msg);
  }

  /**
   * Sends a message to all clients
   * @param message The message that is sent to every client
   */
  public synchronized void sendToAllClients(String message) {
    Collection col = clients.values();
    Iterator it = col.iterator();
    while(it.hasNext()) {
      Client c = (Client)it.next();
      c.sendMsg(message);
    }
  }

  /**
   * Remove a client from the room.
   * This room does not have any direct reference to
   * a client after it has been removed
   * @param userName The name of the client that is being removed
   * @return The client that has just been removed
   */
  public synchronized Client removeClient(String userName) {
    return (Client)clients.remove(userName);
  }

  /**
   * Handles messages from client
   * @param msg the message or command that is being sent
   * @param sender the client who sent the message
   */
  public synchronized void handleMsg(String msg, Client sender) {
    int tag = MessageParser.getTag(msg);
    String theMsg = MessageParser.getMsg(msg);
    System.out.println(msg);
    switch (tag) {
      case Room.LOGIN:
        setUserName(theMsg, sender);
        break;
      case Room.CREATE:
        createGame(theMsg, sender);
        break;
      case Room.JOIN:
        joinGame(theMsg, sender);
        break;
      case Room.LOGOUT:
        logout(sender);
        break;
      case Room.GAME_INFO:
        sendGameList(sender);
        break;
      default:
        System.err.println("Message from " + sender.getUserName() + " not understood:" + msg);
        break;
    }
  }

  /**
   * Checks to see if a given user is in the room.
   * @param userName the user name to search for
   * @return true if the room contains the given user name, otherwise false
   */
  public synchronized boolean containsUser(String userName) {
    return clients.containsKey(userName);
  }

  /**
   * Removes the given game from the game list
   * @param game the game that is being removed
   */
  public synchronized void removeGame(GameRoom game) {
    Iterator it = games.keySet().iterator();
    while(it.hasNext()) {
      String name = (String) it.next();
      if (games.get(name) == game)
        games.remove(name);
    }
  }

  /**
   * Sets the user's name
   * Check to see that the user name is not taken
   * Sends the user a 1 if the user name was changed
   * Sends the user a 0 if the user name was taken
   * @param name The user name
   * @param sender The client whose name needs to be changed
   */
  private synchronized void setUserName(String name, Client sender) {
    boolean inUse = false;
    name = name.toLowerCase();
    if (this.containsUser(name))
      inUse = true;
    Iterator it = games.keySet().iterator();
    while(it.hasNext()) {
      Room r = (Room)games.get(it.next());
      if (r.containsUser(name.toLowerCase()))
        inUse = true;
    }
    if (!inUse) {
      clients.remove(sender.getUserName());
      sender.setUserName(name);
      clients.put(sender.getUserName(), sender);
      sender.sendMsg("1_" + sender.getUserName());
    }
    else { // name in use send 01 and d/c user
      sender.sendMsg("0_1_");
      clients.remove(sender.getUserName());
    }
  }

  private synchronized void logout(Client sender) {
    sender.sendMsg(Room.LOGOUT + "_");
    removeClient(sender.getUserName());
  }

  /**
   * Creates a new game room for the client
   * Replies to client with a 3 if the game is created
   * Replies with a 03 if the game name is already taken
   * @param gameName The name of the game being created
   * @sender The client that is creating the game
   */
  private synchronized void createGame(String gameName, Client sender) {
    // check to see if that game name is already taken
    if (games.containsKey(gameName))
      sender.sendMsg("0_3_"); // error message
    else {
      GameRoom game = new GameRoom(sender, this);
      clients.remove(sender.getUserName());
      games.put(gameName, game);
      // game created !!
      sender.sendMsg("3_");
    }
  }

  /**
   * Puts the client into the given game room
   * Replies to client with 4 if the game was joined
   * Replies with 04 if the game didn't exist or was "in-game"
   * @param gameName The name of the game the client wants to join
   * @param sender The client that is joining the game
   */
  private synchronized void joinGame(String gameName, Client sender) {
    GameRoom game = (GameRoom)games.get(gameName);
    if (game == null || !game.canJoin()) {
      sender.sendMsg("0_4_");
    }
    else { // can't join the game
      game.addNewClient(sender);
      this.removeClient(sender.getUserName());
      sender.sendMsg("4_");
      game.sendToAllOthers(sender, "4_"+sender.getUserName());
    }
  }

  /**
   * Format: 14_game1-#-y_game2-#-n_game3-n
   * Sends the client a list of all the games and their current status
   * @param sender the name of the client that requested the game list
   */
  private synchronized void sendGameList(Client sender) {
    Iterator it = games.keySet().iterator();
    String result = "14_";
    while (it.hasNext()) {
      String gameName = (String) it.next();
      GameRoom game = (GameRoom)games.get(gameName);
      int numPlayers = game.getNumPlayers();
      if (numPlayers != 0) {
        result += gameName.substring(0, gameName.length() - 1) + "-" + numPlayers;
        if (game.canJoin())
          result += "-y_";
        else
          result += "-n_";
      }
      else {
        games.remove(gameName);
      }
    }
    sender.sendMsg(result.substring(0, result.length() - 1));
  }

    //MIKIKO
  // starts the server
   public static void main(String[] args) {
      System.out.println("~~~Openspace~~~\n~~~Version 1.0~~~\n~~~August 13th 2006~~~");
      System.out.println("Server is now running...");
      MainRoom test = new MainRoom();
  }
  //END_MIKIKO
}